#!/usr/bin/env python
# -*- coding: UTF-8 -*-
'''
Created on 21/07/2009
Implement command line details

@author: Salvador de la Puente González
'''

import os
import sys
import time
import fluxcore.fluxcore as fluxcore
from os import path
from optparse import OptionParser, OptionGroup

## Gets the target path taking in count the current working directory and the path
#  given as argument.
#  @param cpath 
def get_target_path(cpath):
    '''Gets the target path taking in count the current working directory and the path
    given as argument'''
    
    target_path = path.normpath(cpath)
    if not path.isabs(target_path):
        target_path = path.join(os.getcwd(), target_path)
        
    return path.normpath(target_path)

## Implements snapshot command
#  @param options options
#  @param target_path path where take a snapshot 
def do_shot(options, target_path):
    '''Implements snapshot command'''
    
    # Check renaming syntax
    if options.rename and not options.move_from:
        raise fluxcore.UsageError("--rename option requires --move-from option as the original path of the file (the name before renaming)")
    
    # By default commentary
    comment = options.comment
    if not comment:
        comment = time.strftime(options.time_pattern, time.localtime())

    # Movement
    if options.move_from:
        source_path = path.normpath(options.move_from)
        if not path.isabs(source_path):
            source_path = path.join(os.getcwd(), source_path)

        fluxcore.snapshot_and_move(target_path, source_path, options.rename, comment=comment)

    # Only a snapshot
    else:
        fluxcore.snapshot_and_move(target_path, comment=comment)
        
## Implements recover command
#  @param options options
#  @param target_path path to recover
def do_recover(options, target_path):
    '''Implements recover command'''
    
    # Check syntax
    if options.timestamp and options.datetime:
        raise fluxcore.UsageError("--timestamp and --datetime options are mutually exclusive")

    # Get the path to the recovered data
    where = path.normpath(options.destination)
    if not path.isabs(where):
        where = path.join(os.getcwd(), where)

    # Get timestamp from a datetime string
    if options.datetime:
        parsed_time = time.strptime(options.datetime, options.time_pattern)
        timestamp = int(time.mktime(parsed_time))

    # Get a real timestamp
    if options.timestamp:
        timestamp = options.timestamp

    # Get current timestamp
    if not options.timestamp and not options.datetime:
        timestamp = int(time.time())

    try:
        fluxcore.recover(target_path, timestamp, where, options.force)
    except IOError, ex:
        raise fluxcore.UsageError(str(ex) + ', use --force option to overwrite it')
    
## Do list command
#  @param options options
#  @param target_path path to take the list of snapshots
def do_list(options, target_path):
    '''Do list command'''
    
    snapshots = fluxcore.get_snapshots(target_path)
    stypes = {'r':'RENAME', 'm':'MOVE', 'b':'BASE', 'i':'INCREMENTAL', 'l':'DIR BASE', 'd':'DELETED', 'h':'HIT'}
    #header = '#\tTimestamp\tLocal time\t\tType\t\tComments\n'
    header = '#\tTimestamp\tType\t\tComments\n'
    sys.stdout.write(header.upper())
    for i in range(len(snapshots)):
        snapshot = snapshots[i]
        
        timestamp = snapshot['timestamp']
        #ltime = time.strftime(options.time_pattern, time.localtime(snapshot['timestamp']))
        comment = snapshot['comment']
        
        #line = '%i\t%s\t%s\t%s\t%s\n' % (i+1, timestamp, ltime, stypes[snapshot['type']], comment)
        line = '%i\t%s\t%s\t\t%s\n' % (i+1, timestamp, stypes[snapshot['type']], comment)
        sys.stdout.write(line.encode('utf-8'))
        
        # Extra info
        if snapshot['related_timeline']:
            related = fluxcore.get_path_from_timeline(snapshot['related_timeline'])
            
        if 'r' == snapshot['type']:
            sys.stdout.write('\t  RENAMED from %s\n\n' % related)
        elif 'm' == snapshot['type']:
            sys.stdout.write('\t  MOVED from %s\n\n' % related)
        elif 'd' == snapshot['type']:
            if snapshot['related_timeline']:
                sys.stdout.write('\t  RELATED TO %s\n\n' % related)

## Encapsulates the command line behavior
#  @param parser the option parser
def do_magick(parser):
    '''Encapsulates the command line behavior'''
    
    (options, args) = parser.parse_args()
    if len(args) != 2:
        parser.error("incorrect number of arguments")

    # Retrieve arguments
    command = args[0]
    target_path = get_target_path(args[1])

    try:
        # Take a snapshos
        if 'shot' == command or 'snapshot' == command:
            do_shot(options, target_path)
            
        # Recover a file
        elif 'recover' == command:
            do_recover(options, target_path)
                    
        # List the snapshots of a path
        elif 'list' == command:
            do_list(options, target_path)
                        
        else:
            raise fluxcore.UsageError('command unknown, please use one of these commands: [snap]shot, recover, list')
            
    except fluxcore.UsageError, ex:
        parser.error(str(ex))


## Point of entry, configures the parser
def main():
    '''Point of entry'''
    ver = '%prog dmc-2'
    usage = 'usage: %prog command [options] path'
    desc = \
"""flux is a simple backup tool that allows to you to capture the state of
any file or directory (we call this 'to take a snapshot') at any moment in order to
restore it in the future. flux is the backup core for HDeLorean but it can run
in standalone mode to use as an usual command line application.
"""
    parser = OptionParser(usage, description=desc, version=ver)
    snapshot_group = OptionGroup(parser,
                                 title='[snap]shot command',
                                 description=\
"""[snap]shot command takes a snapshot of the given path. The options are used to
indicate flux that given path is the destination path of a movement/renaming
operation. In this cases you need to specify the source path.""")

    snapshot_group.add_option('-c', '--comment', type='string', default='',
                      help='add a commentary to the snapshot, the commentary is shown when using list command. If no commentary is given, the current local time is used')
    snapshot_group.add_option('-m', '--move-from', type='string', metavar='SOURCE_PATH',
                      help='notify a movement from SOURCE_PATH and performs a snapshot')
    snapshot_group.add_option('-n', '--rename', action="store_true", default=False,
                      help='requires --move-from option and indicates that the movement is actually a rename')

    recover_group = OptionGroup(parser,
                                title='recover command',
                                description=\
"""recover command restores the indicated path to the last snapshot. If you want to
indicate another moment, use --datetime option and configure the recognition pattern
using --time-pattern option. I.e: 'flux recover -t 2009-01-01 10:30:15' /desired/path""")

    recover_group.add_option('-f', '--force', action='store_true', default=False,
                      help='forces recovered data to overwrite existing files and directories')
    recover_group.add_option('-d', '--destination', metavar='PATH', default='.',
                      help='recovers data into PATH, instead of working directory')
    recover_group.add_option('-s', '--timestamp', type='int', metavar='TIMESTAMP',
                      help='recovers data at given TIMESTAMP. Timestamp must be in Universal Time')
    recover_group.add_option('-p', '--time-pattern', type='string', default='%Y-%m-%d %H:%M:%S', metavar='PATTERN',
                      help="set how to parse a date, default is '%default' that means year-month-day hours:minutes:seconds")
    recover_group.add_option('-t', '--datetime', metavar='DATETIME',
                      help='string indicating date and time of the moment to recover. It will be parsed attending to --time-pattern format string')
    
    list_group = OptionGroup(parser,
                             title='list command',
                             description=\
"""list command show all snapshots of the given path.""")

    parser.add_option_group(snapshot_group)
    parser.add_option_group(recover_group)
    parser.add_option_group(list_group)

    do_magick(parser)
    
if __name__ == '__main__':
    main()
