#!/usr/bin/env python
# -*- coding: UTF-8 -*-

'''
Created on 20/06/2009
Pack access to database

@author: Salvador de la Puente González
'''
from os import path
import settings
from pysqlite2 import dbapi2 as sqlite3

## Pack utilities for timelines table
# TODO: Check the encoding bug. Look for: '# FIXME: Encoding fix'
#
class Timelines(object):
    '''Pack utilities for timelines table'''

    ## Raised when a database object does not exist
    class DoesNotExist(Exception):
        '''Raised when a database object does not exist'''

        ## @param info: additional info about the object that does not exist
        def __init__(self, info):
            Exception.__init__(self)
            self.info = info

        def __str__(self):
            return u'Timeline does not exist (%s)' % str(self.info)

    ## Get a list of timelines
    #  @return a list of timelines
    @staticmethod
    def get_timelines():
        '''Get a list of timelines'''
        
        cursor = DB.cursor()

        timelines = cursor.execute(u"SELECT * FROM timelines ORDER BY name ASC").fetchall()

        cursor.close()

        return timelines

    ## Insert a new timeline based on a path
    # @param cpath path represented by the timeline
    # @return the new id
    # @throws pysqlite2.dbapi2.IntegrityError if the path is already in the database
    @staticmethod
    def add_timeline(cpath):
        '''Insert a new timeline based on a path'''
        
        import time

        # Setup consts
        cpath = cpath.decode('utf-8') # FIXME: Encoding fix
        cpath = path.normpath(cpath)
        now = int(time.time())
        timeline_type = 'f'
        if (path.isdir(cpath)):
            timeline_type = 'd'

        head, final_node = path.split(cpath)

        # Build timelines of paths
        cursor = DB.cursor()

        parent = 0
        folders = head.split('/')[1:]

        insert_mode = False
        for folder in folders:
            # Look for timelines already in the database
            if not insert_mode:
                node = cursor.execute(u"SELECT id FROM timelines WHERE parent=? AND name=?", (parent, folder)).fetchall()
                if node:
                    parent = node[0]['id']

                else:
                    insert_mode = True

            # Insert new timelines (see above: else wont work!)
            if insert_mode:
                cursor.execute(u"INSERT INTO timelines (name, parent, start_time, type) VALUES (?, ?, ?, 'd')", (folder, parent, now))
                parent = cursor.execute(u"SELECT id FROM timelines ORDER BY id DESC LIMIT 1").fetchone()['id']

        # Build timeline of the node
        cursor.execute(u"INSERT INTO timelines (name, parent, start_time, type) VALUES (?, ?, ?, ?)", (final_node, parent, now, timeline_type))

        new_id = cursor.execute(u"SELECT id FROM timelines ORDER BY id DESC LIMIT 1").fetchone()['id']

        DB.commit()
        cursor.close()

        return new_id

    ## Path of a timeline
    # @param t_id id of the timeline
    # @return the path linked to a timeline id
    # @throws Timeline.DoesNotExist: when the id is not within the db
    @staticmethod
    def get_path(t_id):
        '''Path of a timeline'''
        
        folders = []

        current = t_id
        while current:
            timeline = Timelines.get(current)
            folders.append(timeline['name'])
            current = timeline['parent']

        folders.reverse()
        return '/' + '/'.join(folders)

    ## Gets the timeline based on a path
    #  @param cpath path of the timeline to look for
    #  @return id of the timeline linked to the path
    #  @throws Timeline.DoesNotExist: when there is no
    #  timeline linked to the path
    @staticmethod
    def get_timeline(cpath):
        '''Gets the timeline based on a path'''
        cpath = cpath.decode('utf-8') # FIXME: Encoding fix
        cpath = path.normpath(cpath)
        folders = cpath.split('/')[1:]

        cursor = DB.cursor()

        parent = 0
        for folder in folders:
            node = cursor.execute(u"SELECT id FROM timelines WHERE parent=? AND name=?", (parent, folder)).fetchall()
            if node:
                parent = node[0]['id']
            else:
                raise Timelines.DoesNotExist({'id':parent})

        cursor.close()

        return parent

    ## Gets the fields of a timeline 
    #  @param t_id id of the timeline to be retrieved
    #  @return fields of the timeline with that id in a dictionary fashion
    @staticmethod
    def get(t_id):
        '''Gets the fields of a timeline'''
        
        cursor = DB.cursor()

        timeline = cursor.execute(u"SELECT * FROM timelines WHERE id=? LIMIT 1", (t_id,)).fetchone()
        if not timeline:
            raise Timelines.DoesNotExist({'id':t_id})

        cursor.close()

        return timeline

    ## Check database and insert the base timeline if it's needed
    #
    @staticmethod
    def check_timelines():
        '''Check database and insert the base timeline if it's needed'''
        
        cursor = DB.cursor()

        is_empty = 0 == len(cursor.execute(u"SELECT id FROM timelines").fetchall())
        if is_empty:

            # Base timeline
            import time
            now = int(time.time())
            cursor.execute(u"INSERT INTO timelines (id, start_time, name, parent, type) VALUES (0, ?, '/', 0, 'd');" , (now,))
            DB.commit()

        cursor.close()

## Pack utilities for snapshots table
class Snapshots(object):
    '''Pack utilities for snapshots table'''
    
    ## Raised when a database object does not exist
    class DoesNotExist(Exception):
        '''Raised when a database object does not exist'''

        ## @param info additional info about the object that does not exist
        def __init__(self, info):
            Exception.__init__(self)
            self.info = info

        def __str__(self):
            return u'Snapshot does not exist (%s)' % str(self.info)

    ## Get all snapshots for a timeline
    #  @param timeline timeline of the snapshots
    #  @return snapshots of a timeline from the eldest to the newest
    @staticmethod
    def get_snapshots(timeline):
        '''Get all snapshots for a timeline'''
        
        cursor = DB.cursor()

        snapshots = cursor.execute(u"SELECT * FROM snapshots WHERE timeline=? ORDER BY timestamp ASC, id ASC", (timeline,)).fetchall()

        cursor.close()

        return snapshots

    ## Gets last incremental for a timeline, strictly before a given moment
    #  @param timeline timeline of the snapshots
    #  @param timestamp timemark to search before 
    #  @param delta if given, the search look for those snapshots in the
    #  given moment but strictly before delta
    #  @return last incremental for a timeline, strictly before a given moment
    #
    @staticmethod
    def get_incremental(timeline, timestamp, delta=0):
        '''Gets last incremental for a timeline, strictly before a given
        moment'''
        
        cursor = DB.cursor()

        # Delta search
        incremental = None
        if delta:
            incremental = cursor.execute('SELECT * FROM snapshots WHERE timeline=? AND timestamp=? AND id<? ORDER BY timestamp DESC, id DESC LIMIT 1', (timeline, timestamp, delta)).fetchone()

        # Timestamp search
        if not incremental:
            incremental = cursor.execute(u"SELECT * FROM snapshots WHERE timeline=? AND timestamp<? LIMIT 1", (timeline, timestamp)).fetchone()

        cursor.close()

        return incremental
    
    ## Replaces the comment for a snapshot
    #  @param snapshot snapshot to replace its comment
    #  @param new_comment the new comment 
    @staticmethod
    def replace_comment(snapshot, new_comment):
        '''Replaces the comment for a snapshot'''
        cursor = DB.cursor()

        cursor.execute(u"UPDATE snapshots SET comment = ? WHERE id = ?", (new_comment, snapshot))
        DB.commit()

        cursor.close()


    ## Adds a new comment. A comment is a pseudo-snapshot indicating
    #  that something happend.
    #  @param timeline timeline the snapshot belongs to
    #  @param timestamp time of the snapshot
    #  @param comment commentary for the snapshot
    @staticmethod
    def new_comment(timeline, timestamp, comment=''):
        '''Adds a new base snapshot'''
        
        snapshot_type = 'h'

        cursor = DB.cursor()

        cursor.execute(u"INSERT INTO snapshots (timeline, timestamp, type, comment) VALUES (?, ?, ?, ?)", (timeline, timestamp, snapshot_type, comment))
        DB.commit()

        cursor.close()

    ## Adds a new base snapshot
    # @param timeline timeline the snapshot belongs to
    # @param timestamp time of the snapshot
    # @param node_path path where base snapshot file was stored
    # @param comment commentary for the snapshot
    @staticmethod
    def new_base(timeline, timestamp, node_path, comment=''):
        '''Adds a new base snapshot'''
        node_path = node_path.decode('utf-8') # FIXME: Encoding fix
        
        snapshot_type = 'b'

        cursor = DB.cursor()

        cursor.execute(u"INSERT INTO snapshots (timeline, timestamp, type, path, comment) VALUES (?, ?, ?, ?, ?)", (timeline, timestamp, snapshot_type, node_path, comment))
        DB.commit()

        cursor.close()

    ## Adds a new directory base snapshot
    # @param timeline timeline the snapshot belongs to
    # @param timestamp time of the snapshot
    # @param node_path path where base snapshot file was stored
    # @param comment commentary for the snapshot
    @staticmethod
    def new_ls(timeline, timestamp, node_path, comment=''):
        '''Adds a new directory base snapshot'''
        node_path = node_path.decode('utf-8') # FIXME: Encoding fix
        
        snapshot_type = 'l'

        cursor = DB.cursor()

        cursor.execute(u"INSERT INTO snapshots (timeline, timestamp, type, path, comment) VALUES (?, ?, ?, ?, ?)", (timeline, timestamp, snapshot_type, node_path, comment))
        DB.commit()

        cursor.close()

    ## Adds a new incremental snapshot
    # @param timeline timeline the snapshot belongs to
    # @param timestamp time of the snapshot
    # @param node_path path where incremental snapshot file was stored
    # @param comment commentary for the snapshot
    @staticmethod
    def new_incremental(timeline, timestamp, node_path, comment=''):
        '''Adds a new incremental snapshot'''
        node_path = node_path.decode('utf-8') # FIXME: Encoding fix
        
        snapshot_type = 'i'

        cursor = DB.cursor()

        cursor.execute(u"INSERT INTO snapshots (timeline, timestamp, type, path, comment) VALUES (?, ?, ?, ?, ?)", (timeline, timestamp, snapshot_type, node_path, comment))
        DB.commit()

        cursor.close()

    ## Adds a new delete snapshot
    # @param timeline timeline the snapshot belongs to
    # @param timestamp time of the snapshot
    # @param related_to optional, if exists, it indicates the deletion is related to a move operation.
    # Then related_to is the id of the destination timeline
    # @param comment commentary for the snapshot
    @staticmethod
    def new_delete(timeline, timestamp, related_to=None, comment=''):
        '''Adds a new delete snapshot'''
        
        snapshot_type = 'd'

        cursor = DB.cursor()

        if related_to is None:
            cursor.execute(u"INSERT INTO snapshots (timeline, timestamp, type, comment) VALUES (?, ?, ?, ?)", (timeline, timestamp, snapshot_type, comment))
        else:
            cursor.execute(u"INSERT INTO snapshots (timeline, timestamp, type, related_timeline, comment) VALUES (?, ?, ?, ?, ?)", (timeline, timestamp, snapshot_type, related_to, comment))
        DB.commit()

        cursor.close()

    ## Adds a new rename snapshot
    # @param old_timeline timeline of the previous name
    # @param new_timeline timeline of the new name
    # @param timestamp time of the snapshot
    # @param comment commentary for the snapshot
    @staticmethod
    def new_rename(old_timeline, new_timeline, timestamp, comment=''):
        '''Adds a new rename snapshot'''
        
        snapshot_type = 'r'

        cursor = DB.cursor()

        cursor.execute(u"INSERT INTO snapshots (timeline, timestamp, type, related_timeline, comment) VALUES (?, ?, ?, ?, ?)", (new_timeline, timestamp, snapshot_type, old_timeline, comment))
        DB.commit()

        cursor.close()

    ## Adds a new move snapshot
    # @param old_timeline timeline of the previous name
    # @param new_timeline timeline of the new name
    # @param timestamp time of the snapshot
    # @param comment commentary for the snapshot
    @staticmethod
    def new_move(old_timeline, new_timeline, timestamp, comment=''):
        '''Adds a new move snapshot'''
        
        snapshot_type = 'm'

        cursor = DB.cursor()

        cursor.execute(u"INSERT INTO snapshots (timeline, timestamp, type, related_timeline, comment) VALUES (?, ?, ?, ?, ?)", (new_timeline, timestamp, snapshot_type, old_timeline, comment))
        DB.commit()

        cursor.close()

## Packs utilities to manage Bases view
class Bases(object):
    '''Packs utilities to manage Bases view'''

    ## Raised when a database object does not exist
    class DoesNotExist(Exception):
        '''Raised when a database object does not exist'''

        ## @param info additional info about the object that does not exist
        def __init__(self, info):
            Exception.__init__(self)
            self.info = info

        def __str__(self):
            return u'Base snapshot does not exist (%s)' % str(self.info)

    ## Retrieve all fields of the nearest (but strict before) base
    #  to certain moment inside a timeline
    #  @param timeline timeline where perform the search
    #  @param timestamp reference moment in time
    #  @param delta optional, if given, get_base returns the nearest
    #  base with the exact timestamp but lower delta
    #  @return a dictionary with these fields or None if it does not exist
    @staticmethod
    def get_base(timeline, timestamp, delta=0):
        '''Retrieve all fields of the nearest (but strict before) base to
        certain moment inside a timeline.'''

        cursor = DB.cursor()

        base = None

        # Delta search
        if delta:
            base = cursor.execute('SELECT * FROM bases WHERE timeline=? AND timestamp=? AND id<? ORDER BY timestamp DESC, id DESC LIMIT 1', (timeline, timestamp, delta)).fetchone()

        # Timestamp search
        if not base:
            base = cursor.execute('SELECT * FROM bases WHERE timeline=? AND timestamp<? ORDER BY timestamp DESC, id DESC LIMIT 1', (timeline, timestamp)).fetchone()

        cursor.close()

        return base

# Configure database    
DB_PATH = path.join(settings.BACKUP_PATH, 'TIME.db')

# Create database
if not path.exists(DB_PATH):
    import shutil
    import os
    EMPTY_PATH = os.path.join(os.path.split(__file__)[0], 'TIME.empty.db')
    shutil.copy(EMPTY_PATH, DB_PATH)

DB = sqlite3.connect(DB_PATH)
DB.row_factory = sqlite3.Row    # Enables by-name column access

Timelines.check_timelines()
