#!/usr/bin/env python
# -*- coding: UTF-8 -*-

'''
Created on 22/06/2009
Define the binary difference builder interface

@author: Salvador de la Puente González
'''
import settings

## Interface to build binary differences
class IBinaryDiffBuilder(object):
    '''Interface to build binary differences'''
    
    ## Builds a binary difference between two files and save it to another file.
    #  This function must compute a patch between \em base and \em current so if the patch
    #  is apply over \em base results in \em current.
    #  @param base path to the file with the original state
    #  @param current path to the file with the current state
    #  @param patch path to the result patch
    #  @return level of compression of the patch
    def make_patch(self, base, current, patch):
        '''Builds a binary difference between two files and save it to another file.'''
        
        import sys
        sys.stderr.write('Method not implemented yet')
    
    ## Applies a binary difference to a file.
    #  @param base path to the file on which the patch will be applied
    #  @param patch path to the patch
    #  @param dst path to the destination file
    def apply_patch(self, base, patch, dst):
        '''Applies a binary difference to a file'''
        
        import sys
        sys.stderr.write('Method not implemented yet')
    
## Uses xdelta3 utility to perform the binary differences
class Xdelta3DiffBuilder(IBinaryDiffBuilder):
    '''Uses xdelta3 utility to perform the binary differences'''
    
    ## Raised if xdelta3 utility is not available
    class XDeltaNotAvailable(Exception):
        '''Raised if xdelta3 utility is not available'''
        
        def __str__(self):
            return 'xdelta3 utility is not available'
        
    ## Raised if xdelta3 return with erros
    class XDeltaError(Exception):
        '''Raised if xdelta3 return with erros'''
        
        def __init__(self, code, info):
            Exception.__init__(self)
            self.code = code
            self.info = info
        
        def __str__(self):
            return 'xdelta3 return with code %i: %s' % (self.code, self.info) 
    
    ## Builds a binary difference between two files and save it to another file.
    #  This function must compute a patch between \em base and \em current so if the patch
    #  is apply over \em base results in \em current.
    #  @param base path to the file with the original state
    #  @param current path to the file with the current state
    #  @param patch path to the result patch
    #  @return level of compression of the patch
    def make_patch(self, base, current, patch):
        '''Builds a binary difference between two files and save it to another file.
        This function must compute a patch between \em base and \em current so if the patch
        is apply over \em base results in \em current.'''
        import subprocess
        
        command = ['xdelta3', 'encode', settings.LEVEL_OF_COMPRESSION, '-v', '-f', '-s', base, current, patch]
        
        try:
            process = subprocess.Popen(command, stderr=subprocess.STDOUT, stdout=subprocess.PIPE)
        except OSError:
            raise Xdelta3DiffBuilder.XDeltaNotAvailable()
        
        output = process.communicate()[0]
        if process.returncode:
            raise Xdelta3DiffBuilder.XDeltaError(process.returncode, output)
        
        return Xdelta3DiffBuilder.__parse_compression(output)
    
    ## Applies a binary difference to a file.
    #  @param base path to the file on which the patch will be applied
    #  @param patch path to the patch
    #  @param dst path to the destination file
    def apply_patch(self, base, patch, dst):
        '''Applies a binary difference to a file'''
        
        import subprocess
        
        command = ['xdelta3', 'decode', '-f', '-s', base, patch, dst]
        
        try:
            process = subprocess.Popen(command, stderr=subprocess.STDOUT, stdout=subprocess.PIPE)
        except OSError:
            raise Xdelta3DiffBuilder.XDeltaNotAvailable()
        
        output = process.communicate()[0]
        if process.returncode:
            raise Xdelta3DiffBuilder.XDeltaError(process.returncode, output)
        
    ## Extracts compression for a verbose output of xdelta
    #  @param output verbose output of xdelta3
    @staticmethod
    def __parse_compression(output):
        '''Extracts compression for a verbose output of xdelta'''
        
        compression_line = output.splitlines()[-1]
        compression_token = compression_line.split(' ')[-1]
        compression_level = float(compression_token.strip()[1:-2]) / 100
        return compression_level