#!/usr/bin/env python
# -*- coding: UTF-8 -*-

'''
Created on 21/06/2009
The main file with implementation of flux capabilities

@author: Salvador de la Puente González
'''
import settings
import db
import os
import shutil
import time
import sys
try:
    import cPickle as pickle
except ImportError:
    import pickle

from os import path
from diff import Xdelta3DiffBuilder as xdelta

## Represent errors in usage
class UsageError(Exception):
    '''Represent errors in usage'''

    def __init__(self, info):
        Exception.__init__(self)
        self.info = info

    def __str__(self):
        return u"Invalid usage of flux >> %s" % self.info

## Recover operation is impossible
class RecoverImpossible(Exception):
    '''Recover operation is impossible'''

    def __init__(self, info):
        Exception.__init__(self)
        self.info = info

    def __str__(self):
        return u"Recover operation is impossible: %s" % self.info

## Gets a timeline for a path and optionally, creates one if it does not exist
#  @param p the path of the timeline
#  @param add_new if True and the timeline for the path does not exist, this option make new one.
#  @return the timeline or None
def __get_timeline(p, add_new=False):
    '''Gets a timeline for a path and optionally, creates one if it does not exist'''
    
    try:
        timeline = db.Timelines.get_timeline(p)

    except db.Timelines.DoesNotExist:
        if add_new:
            timeline = db.Timelines.add_timeline(p)
        else:
            timeline = None

    return timeline

## Gets the base snapshot of a file timeline nearest to a moment given and following
#  the redirections due to renames or movements.
#  @param timeline the timeline to get the base
#  @param timestamp moment given
#  @return the base fields (access them in a dictionary fashion) 
def __get_base(timeline, timestamp):
    '''Gets the base snapshot of a file timeline nearest to a moment given and following
    the redirections due to renames or movements.'''
    
    base_type = 'b'     # Actual base type for files
    if db.Timelines.get(timeline)['type'] != 'f':
        base_type = 'l' # Actual base type for directories

    base = db.Bases.get_base(timeline, timestamp + 1)
    while base and base['type'] != base_type:
        delta = base['id']
        timestamp = base['timestamp']
        timeline = base['related_timeline']

        # Extra condition to exit:
        # If it is not possible to follow the timeline, there is no base.
        if timeline is None:
            return None

        base = db.Bases.get_base(timeline, timestamp, delta)

    return base

## Gets the incremental snapshot of a file timeline nearest to a moment
#  given and following the redirections due to renames or movements.
#  @param timeline the timeline to get the base
#  @param timestamp moment given
#  @return the incremental fields (access them in a dictionary fashion) 
def __get_incremental(timeline, timestamp):
    '''Gets the incremental snapshot of a file timeline nearest to a moment
    given and following the redirections due to renames or movements.'''
    
    incremental = db.Snapshots.get_incremental(timeline, timestamp + 1)
    while incremental and incremental['type'] != 'i':
        delta = incremental['id']
        timestamp = incremental['timestamp']
        timeline = incremental['related_timeline']

        # Extra condition to exit:
        # If it is not possible to follow the timeline, there is no incremental.
        if not timeline:
            return None

        incremental = db.Snapshots.get_incremental(timeline, timestamp, delta)

    return incremental

## Takes a new snapshot from a directory
#  @param cpath path of the directory
#  @param timestamp moment of the snapshot, if ignored, now is taken
#  @return a set with the differences between the last contents and current contents
#  @param comment if given, adds a commentary to the snapshot
def __new_dir_snapshot(cpath, timestamp=None, comment=''):
    '''Takes a new snapshot from a directory'''
    
    if not path.isdir(cpath):
        raise UsageError(u"'%s' is not a directory" % cpath)

    # Get the timeline
    timeline = __get_timeline(cpath, True)

    # Get the base 
    # (for directories, base is a serialized version of a set with its contents)
    
    if not timestamp:
        timestamp = int(time.time())
    gmtime = time.gmtime(timestamp)
    base = __get_base(timeline, timestamp)

    # Build destination path
    tpath = time.strftime("%Y-%m-%d/%H:%M/%S", gmtime)
    head, name = path.split(cpath)
    bk_path = path.join(settings.BACKUP_PATH, tpath, head[1:], name)
    bk_path = path.join(bk_path, 'ls')

    # Get new set of contents
    flist = set(os.listdir(cpath))

    base_needed = True
    deleted = set()
    added = set()
    if base:
        # Get old set
        try:
            cfile = open(base['path'], 'rb')
            old_list = pickle.load(cfile)
        finally:
            if cfile:
                cfile.close()

        # Compute deleted
        deleted = old_list - flist
        added = flist - old_list
        del old_list  # Free old_ls

        # No new bases needed
        if not deleted and not added:
            base_needed = False

    # New base is needed
    if base_needed:
        header = path.split(bk_path)[0]
        if not path.exists(header):
            os.makedirs(header)

        try:
            cfile = open(bk_path, 'w+b')
            pickle.dump(flist, cfile, pickle.HIGHEST_PROTOCOL)
        finally:
            if cfile:
                cfile.close()

        db.Snapshots.new_ls(timeline, timestamp, bk_path, comment=comment)
        
    # Nothing is needed so we record a comment
    else:
        db.Snapshots.new_comment(timeline, timestamp, comment=comment)

    return deleted, added

## Takes a new snapshot from a file
#  @param cpath path of the file
#  @param timestamp moment of the snapshot, if ignored, now is taken
#  @return tuple (is_base, path_to_backup). Check is_base to determine if path is a patch or an integral backup.
#  Note that the integral backup can be a hardlink to a previous snapshot so check this manually please.
#  @param comment if given, adds a commentary to the snapshot  
def __new_file_snapshot(cpath, timestamp=None, comment=''):
    '''Takes a new snapshot from a file'''
    
    if not path.isfile(cpath):
        raise UsageError(u"'%s' is not a file" % cpath)

    # Get the timeline
    timeline = __get_timeline(cpath, True)

    # Get the base
    
    if timestamp is None:
        timestamp = int(time.time())
    gmtime = time.gmtime(timestamp)
    base = __get_base(timeline, timestamp)

    # Build destination path
    tpath = time.strftime("%Y-%m-%d/%H:%M/%S", gmtime)
    head, name = path.split(cpath)
    bk_path = path.join(settings.BACKUP_PATH, tpath, head[1:])
    if not path.exists(bk_path):
        os.makedirs(bk_path)
    bk_path = path.join(bk_path, name + '.bk')

    base_needed = True
    if base:
        # There is a base -> make a patch
        dbuilder = xdelta()
        compression_level = dbuilder.make_patch(base['path'], cpath, bk_path)

        # Over the limit of compression -> delete incremental and force new base
        if compression_level > settings.COMPRESSION_THRESHOLD:
            os.unlink(bk_path)
        else:
            db.Snapshots.new_incremental(timeline, timestamp, bk_path, comment=comment)
            base_needed = False

    # New base is needed
    if base_needed:
        shutil.copy(cpath, bk_path)
        db.Snapshots.new_base(timeline, timestamp, bk_path, comment=comment)

    return base_needed, bk_path

## Notifies the deletion of a file
#  @param cpath path to notify
#  @param timestamp moment of deletion, if ignored, now is taken
#  @param related_to a timeline id that the current timeline is related to because, i. e.
#  is the result of a file movement.
#  @param comment if given, adds a commentary to the snapshot
#  @return the timeline updated
def __notify_del(cpath, timestamp=None, related_to=None, comment=''):
    '''Notifies the deletion of a file'''
    
    # Get the timeline
    timeline = __get_timeline(cpath)

    # Check for None. Timeline Id can be 0 if it is the system root
    if timeline is not None:
        # Get timestamps
        
        if timestamp is None:
            timestamp = int(time.time())

        # Save the deletion
        db.Snapshots.new_delete(timeline, timestamp, related_to, comment=comment)

    return timeline

## Notifies the movement of a file
#  @param path1 source path
#  @param path2 destination path
#  @param timestamp moment of movement, if ignored, now is taken
#  @param rename optional, if True, notify a rename instead a movement
#  @return the timeline updated
#  @param comment if given, adds a commentary to the snapshot
def __notify_move(path1, path2, timestamp=None, rename=False, comment=''):
    '''Notifies the movement of a file'''
    
    if not path.exists(path2):
        raise UsageError(u"'%s' must exist" % path2)

    # Get the timeline of the destination (add if necessary)
    ts2 = __get_timeline(path2, True)

    # Get timestamps
    
    if timestamp is None:
        timestamp = int(time.time())

    # Notify deletion of p1
    ts1 = __notify_del(path1, timestamp, ts2)

    # Notify movement (or rename)
    if rename:
        db.Snapshots.new_rename(ts1, ts2, timestamp, comment=comment)
    else:
        db.Snapshots.new_move(ts1, ts2, timestamp, comment=comment)

    return ts1

## Gets the base and patch to be applied to it in order to recover
#  the timeline at the given moment
#  @param timeline path to get its base and patch
#  @param timestamp moment
#  @return a tuple of snapshots <base, patch>. Patch can be None if it is
#  not required. 
def __get_base_and_patch(timeline, timestamp):
    '''Gets the base and patch to be applied to it in order to recover
    the timeline at the given moment'''
    
    base = __get_base(timeline, timestamp)
    patch = __get_incremental(timeline, timestamp)

    # No base -> no patch
    if not base:
        return None, None

    # Patch cannot be after base
    if patch and patch['id'] < base['id']:
        patch = None

    return base, patch

## Recovers a file.
#  @param timeline timeline fields (in a dictionary fashion, retrieve them by using
#  Timeline.get_timeline method)
#  @param timestamp moment at it is wanted to recover the path
#  @param where location to put the recovered file
#  @param force_overwrite if True and path to recovered data already exists,
#  it is overwrote. 
#  @return path to the recovered data
def __recover_file(timeline, timestamp, where, force_overwrite=False):
    '''Recovers a file'''
    
    t_id = timeline['id']

    if timeline['type'] != 'f':
        raise UsageError("timeline '%i' is not a file" % t_id)

    path_to_recovered_data = path.join(where, timeline['name'])

    # Check overwrite
    if not force_overwrite and path.exists(path_to_recovered_data):
        raise IOError("'%s' already exists" % path_to_recovered_data)

    base, patch = __get_base_and_patch(t_id, timestamp)

    # No base -> recover impossible
    if not base:
        raise RecoverImpossible("timeline '%i' has no base snapshot" % t_id)

    # Just a base -> copy it to recover
    if not patch:
        shutil.copyfile(base['path'], path_to_recovered_data)

    # Base and patch -> apply patch to base and restore
    else:
        dbuilder = xdelta()
        dbuilder.apply_patch(base['path'], patch['path'], path_to_recovered_data)

    return path_to_recovered_data

## "Recovers" a directory. Really, this method does not recover nothing but create
#  the directory and retrive the directory contents at the moment given.
#  @param timeline timeline fields (in a dictionary fashion, retrieve them by using
#  Timeline.get_timeline method).
#  @param timestamp moment at it is wanted to recover the path
#  @param where location to put the recovered file
#  @param force_overwrite if True and path to recovered data already exists,
#  it is overwrote. 
#  @return a set with the contents of the directory at the given moment
def __recover_dir(timeline, timestamp, where, force_overwrite=False):
    '''"Recovers" a directory. Really, this method does not recover nothing but create
    the directory and retrive the directory contents at the moment given.'''
    
    t_id = timeline['id']

    if timeline['type'] != 'd':
        raise UsageError("timeline '%i' is not a directory" % t_id)

    path_to_recovered_data = path.join(where, timeline['name'])

    # Check overwrite
    if not force_overwrite and path.exists(path_to_recovered_data):
        raise IOError("'%s' already exists" % path_to_recovered_data)

    base = __get_base(timeline['id'], timestamp)

    # No base -> recover impossible
    if not base:
        raise RecoverImpossible("timline '%i' has no directory contents" % t_id)

    if not path.exists(path_to_recovered_data):
        os.makedirs(path_to_recovered_data)

    try:
        cfile = open(base['path'], 'rb')
        contents = pickle.load(cfile)
    finally:
        if cfile:
            cfile.close()

    return contents

## Retrieves data about the snapshots of a timeline
#  @param cpath path to retrieve its snapshots
#  @return a list of fields (in a dictionary fashion) with all information about snapshots
def get_snapshots(cpath):
    '''Retrieves data about the snapshots of a timeline'''
    
    try:
        t_id = db.Timelines.get_timeline(cpath)
    except db.Timelines.DoesNotExist:
        raise UsageError("It is not possible to retrieve '%s' snapshots because the path is not tracked" % cpath)

    snapshots = db.Snapshots.get_snapshots(t_id)
                
    return snapshots

## Returns True if given path is tracked
def is_tracked(cpath):
    '''Returns True if given path is tracked'''
    
    try:
        db.Timelines.get_timeline(cpath)
        return True
    except db.Timelines.DoesNotExist:
        return False

## Makes a recursive snapshot over a directory and optionally, notify a 
#  movement from some source.
#  @param cpath path to make the snapshot
#  @param source if given, location where current location was supposed to be before
#  @param rename if true, indicate that the movement is a renaming
#  movement.
#  @param comment if given, adds a commentary to the snapshot. The same comentary
#  is given to all sub-snapshots required by this snapshot  
def snapshot_and_move(cpath, source=None, rename=False, comment=''):
    '''Makes a recursive snapshot over a directory and optionally, notify a 
    movement from some source.'''

    ## Recursive snapshot method
    def __snapshot_and_move(cpath, timestamp, source=None, rename=False, comment=''):
        '''Recursive snapshot method'''
        
        # Contemplate WHITE & BLACK LISTS
        current_parent_path, fname = path.split(cpath)
        if (fname in settings.WHITELIST) or (fname not in settings.BLACKLIST):
        
            # Notify movement if required
            if source is not None:
                # Notify alteration of directory contents
                source_parent_path = path.split(source)[0]
                __new_dir_snapshot(current_parent_path, comment=comment)
                __new_dir_snapshot(source_parent_path, comment=comment)
    
                __notify_move(source, cpath, timestamp, rename, comment=comment)
    
            # Base case -> regular file
            if path.isfile(cpath):
                __new_file_snapshot(cpath, timestamp, comment=comment)
    
            # Recursive case -> folder
            if path.isdir(cpath):
                # Retrieve deleted items
                deleted = __new_dir_snapshot(cpath, timestamp, comment=comment)[0]
    
                # Notify deleted items
                for del_entry in deleted:
                    path_to_entry = path.join(cpath, del_entry)
                    __notify_del(path_to_entry, timestamp)
    
                # Current content
                flist = set(os.listdir(cpath))
                for entry in flist:
                    source_to_entry = None
                    if source is not None:
                        path.join(source, entry)
    
                    path_to_entry = path.join(cpath, entry)
                    __snapshot_and_move(path_to_entry, timestamp, source_to_entry, rename, comment=comment)
    
            # TODO: What happen with symbolic link? Check it!

    
    now = int(time.time())
    __snapshot_and_move(cpath, now, source, rename, comment=comment)

## Recovers a file or directory in a given moment
#  @param p path to be recovered
#  @param timestamp moment at it is wanted to recover the path
#  @param where location to put the recovered file
#  @param force_overwrite if True and path to recovered data already exists,
#  it is overwrote.
#  @return path to the recovered data
def recover(cpath, timestamp, where, force_overwrite=False):
    '''Recovers a file or directory in a given moment'''

    ## Wrapped recursive method
    def __recover(cpath, timestamp, where, force_overwrite, errors):
        '''Wrapped recursive method'''
        
        try:
            timeline = db.Timelines.get(db.Timelines.get_timeline(cpath))
        except db.Timelines.DoesNotExist, ex:
            raise RecoverImpossible("'%s' cannot be recovered because it was not tracked before." % cpath)

        # A file
        if 'f' == timeline['type']:
            __recover_file(timeline, timestamp, where, force_overwrite)

        # A directory
        if 'd' == timeline['type']:
            contents = __recover_dir(timeline, timestamp, where, force_overwrite)

            for item in contents:
                path_to_recover = path.join(cpath, item)
                new_where = path.join(where, timeline['name'])
                try:
                    recover(path_to_recover, timestamp, new_where, force_overwrite)
                except RecoverImpossible, ex:
                    errors.append(ex)

        # What happen with hard links?

    errors = []
    __recover(cpath, timestamp, where, force_overwrite, errors)
    sys.stderr.writelines(errors)

## Gets the path of the timeline given    
def get_path_from_timeline(t_id):
    '''Gets the path of the timeline given'''
    return db.Timelines.get_path(t_id)

## Replaces the comment for a snapshot
def replaces_comment(snapshot, new_comment):
    '''Replaces the comment for a snapshot'''
    return db.Snapshots.replace_comment(snapshot, new_comment)