#!/usr/bin/env python
# -*- coding: UTF-8 -*-

'''
Created on 03/10/2009

@author: Salvador de la Puente González
'''

import gtk
import os
import time
import dircache
import fluxcore.settings as settings
import fluxcore.fluxcore as fluxcore
from os import path

## Class with signal definitions for ui files
class FluxGUI:
    '''Class with signal definitions for ui files'''
    
    def __init__(self):
        # Build interface
        self.builder = gtk.Builder()
        self.builder.set_translation_domain('fluxgui')
        try:
            self.builder.add_from_file('ui/main_window.ui')
        except:
            self.builder.add_from_file('/usr/share/fluxgui/main_window.ui')
            
        self.builder.connect_signals(self)
        
        # Associate GUI with class' fields
        self.window = self.builder.get_object('main_window')
        
        # Path entry with completition
        self.cpath = self.builder.get_object('current_path')
        completion = gtk.EntryCompletion()
        completion.set_model(gtk.ListStore(str))
        completion.set_text_column(0)
        completion.set_inline_completion(True)
        self.cpath.set_completion(completion)
        
        self.choices = self.cpath.get_completion().get_model()
        
        self.snapshot_list = self.builder.get_object('snapshot_list')
        self.snapshot_list.clear()
        self.snapshot_list_view = self.builder.get_object('snapshot_list_view')
        
        # Recover area
        recover_path = settings.RECOVER_PATH
        if not os.path.exists(recover_path):
            os.makedirs(recover_path)
            
        self.dir_chooser = self.builder.get_object('dir_chooser')
        self.recover_button = self.builder.get_object('recover_button')
        self.check_overwrite = self.builder.get_object('check_overwrite')
            
        self.dir_chooser.set_current_folder(recover_path)
        
        # Info line
        self.info = self.builder.get_object('info')
        
        # Start options
        self.cpath.set_text(os.getcwd())
        self.list_snapshots(self.get_target_path())
        
    ## Behavior when no snapshot is selected
    def on_snapshot_list_view_unselect_all(self, widget):
        '''Behavior when no snapshot is selected'''
        self.recover_button.set_sensitive(False)
        
    ## Behavior when recover button is clicked
    def on_recover_button_clicked(self, widget):
        '''Behavior when recover button is clicked'''        
        iterator = self.get_selected_snapshot()
        if iterator:
            target_path = self.get_target_path()
            timestamp = self.snapshot_list.get_value(iterator, 2)
            where = self.dir_chooser.get_filename()
            
            self.recover_snapshot(target_path, timestamp, where)
    
    ## Behavior when a row is selected by double click
    def on_snapshot_list_view_row_activated(self, widget, path, view_column):
        '''Behavior when a row is selected by double click'''
        self.on_recover_button_clicked(widget)
        
    ## Behavior when list button is clicked
    def on_button_list_clicked(self, widget):
        self.list_snapshots(self.get_target_path())
        
    ## Behavior to change the comment
    def on_comment_cell_render_edited(self, widget, path, new_text):
        '''Behavior to change the comment'''
        #TODO: When we change a comment of just one snapshot (by id)? Must we change all comments at the same time (by timestamp)?
        # This is because we can make just one snapshot at time, so all the snapshots with the same timestamps belongs to the same "set"
        
        new_text = new_text.decode('utf-8')
        iterator = self.get_selected_snapshot()
        if iterator:
            # Gets the snapshot and change its comment
            snapshot = self.snapshot_list.get_value(iterator, 0)
            fluxcore.replaces_comment(snapshot, new_text)
            
            # Update view
            self.snapshot_list.set_value(iterator, 4, new_text)
            
    ## Calculate autocompletion
    def on_current_path_changed(self, widget):
        # Reset choices
        model = self.choices
        model.clear()
        
        text = self.cpath.get_text()
        if text:
            path = os.path.dirname(text)
            if os.path.isdir(path):
                start = os.path.basename(text)
                files = dircache.listdir(path)
                for file in files:
                    if file.startswith(start):
                        model.append([os.path.join(path, file)])
                    
    ## Bahvior when a key is pressed during edition of the path
    def on_current_path_key_press(self, widget, event):
        '''Bahvior when a key is pressed during edition of the path'''
        if "Return" == gtk.gdk.keyval_name(event.keyval):
            self.recover_button.set_sensitive(False)
            self.list_snapshots(self.get_target_path())
                    
    
    ## Behavior when user selects a snapshot
    def on_snapshot_list_view_cursor_changed(self, widget):
        '''Behavior when user selects a snapshot'''
        iterator = self.get_selected_snapshot()
        if iterator:
            self.recover_button.set_sensitive(True)
            self.info.set_text(self.snapshot_list.get_value(iterator, 5))
        
    
    ## Behavior when shot button is clicked
    def on_button_shot_clicked(self, widget):
        '''Behavior when shot button is clicked'''
        target_path = self.get_target_path()
        self.do_snapshot(target_path)
        self.list_snapshots(target_path)
        
    ## Behavior to end the application
    def on_main_window_destroy(self, widget):
        '''Behavior to end the application'''
        gtk.main_quit()
        
    ## Get iterator of selected snapshot
    def get_selected_snapshot(self):
        '''Get selected snapshot id'''
        iterator = self.snapshot_list_view.get_selection().get_selected()[1]        
        return iterator
        
    ## Make a snapshot
    def do_snapshot (self, target_path):
        '''Make a snapshot'''        
        if target_path:
            comment = time.strftime('%Y-%m-%d %H:%M:%S', time.localtime())
            fluxcore.snapshot_and_move(target_path, comment=comment)
            
        self.list_snapshots(target_path)
        
    ## Recover a snapshot, always overwrite contents
    def recover_snapshot (self, target_path, timestamp, where):
        '''Recover a snapshot, always overwrite contents'''
        
        try:
            fluxcore.recover(target_path, timestamp, where, self.check_overwrite.get_active())
            self.info.set_text('Data recovered to %s' % where)
                
        except IOError, ex:
            self.info.set_text(str(ex))
            
    ## List the snapshots of the given path
    def list_snapshots(self, target_path):
        '''List the snapshots of the given path'''
        
        try:
            snapshots = fluxcore.get_snapshots(target_path)[::-1]
            stypes = {'r':'RENAME', 'm':'MOVE', 'b':'BASE', 'i':'INCREMENTAL', 'l':'DIR BASE', 'd':'DELETED', 'h':'HIT'}
            
            self.snapshot_list.clear()
            for i in range(len(snapshots)):
                count = i+1
                snapshot = snapshots[i]
                identifier = snapshot['id']            
                timestamp = snapshot['timestamp']
                stype = stypes[snapshot['type']]
                comment = snapshot['comment']
                
                # Extra info
                extra = 'No extra info'
                if snapshot['related_timeline']:
                    related = fluxcore.get_path_from_timeline(snapshot['related_timeline'])
                    
                if 'r' == snapshot['type']:
                    extra = 'renamed from %s' % related
                elif 'm' == snapshot['type']:
                    extra = 'moved from %s' % related
                elif 'd' == snapshot['type']:
                    if snapshot['related_timeline']:
                        extra = 'related to %s' % related
                
                self.snapshot_list.append([identifier, count, timestamp, stype, comment, extra])
                        
            # Info tag
            if len(snapshots):
                self.info.set_text('%i snapshots loaded!' % len(snapshots))
            else:
                self.info.set_text('No snapshots for the current path!')
        
        except fluxcore.UsageError, ex:
            self.snapshot_list.clear()
            self.info.set_text(ex.info)
                    
    ## Gets the current target path
    def get_target_path(self):
        return self.cpath.get_text()

## Main
def main():
    '''Main'''
    gui = FluxGUI()
    gui.window.show()
    gtk.main()

if __name__ == '__main__':
    main()